@@ -7,12 +7,13 @@ from lightwood.config.config import CONFIG
 
 class ChannelPoolAdaptiveAvg1d(torch.nn.AdaptiveAvgPool1d):
     def forward(self, input):
-        n, c, _, _ = input.size()
-        input = input.view(n,c,1).permute(0,2,1)
-        pooled =  torch.nn.functional.adaptive_avg_pool1d(input, self.output_size)
-        _, _, c = pooled.size()
-        pooled = pooled.permute(0,2,1)
-        return pooled.view(n,c)
+        with LightwoodAutocast():
+            n, c, _, _ = input.size()
+            input = input.view(n,c,1).permute(0,2,1)
+            pooled =  torch.nn.functional.adaptive_avg_pool1d(input, self.output_size)
+            _, _, c = pooled.size()
+            pooled = pooled.permute(0,2,1)
+            return pooled.view(n,c)
 
 class Img2Vec(nn.Module):
 
@@ -35,16 +36,17 @@ class Img2Vec(nn.Module):
         return self
 
     def forward(self, image, batch=True):
-        embedding = self.model(image.to(self.device))
+        with LightwoodAutocast():
+            embedding = self.model(image.to(self.device))
 
-        if self.model_name in ('resnext-50-small'):
-            if batch:
-                return embedding
-            return embedding[0, :]
-        else:
-            if batch:
-                return embedding[:, :, 0, 0]
-            return embedding[0, :, 0, 0]
+            if self.model_name in ('resnext-50-small'):
+                if batch:
+                    return embedding
+                return embedding[0, :]
+            else:
+                if batch:
+                    return embedding[:, :, 0, 0]
+                return embedding[0, :, 0, 0]
 
 
     def _get_model(self):
