@@ -1,27 +1,29 @@
-from __future__ import unicode_literals, print_function, division
-
 from lightwood.encoders.time_series.helpers.rnn_helpers import *
+from lightwood.encoders.time_series.helpers.transformer_helpers import *
 from lightwood.helpers.torch import LightwoodAutocast
 from lightwood.encoders.encoder_base import BaseEncoder
 from lightwood.encoders.datetime import DatetimeEncoder
 from lightwood.helpers.device import get_devices
 
-import numpy as np
 import torch
 import torch.nn as nn
 from torch import optim
+from math import gcd
 
 
-class RnnEncoder(BaseEncoder):
+class TimeSeriesEncoder(BaseEncoder):
 
-    def __init__(self, encoded_vector_size=100, train_iters=100, stop_on_error=0.01, learning_rate=0.01,
-                 is_target=False, ts_n_dims=1):
+    def __init__(self, encoded_vector_size=128, train_iters=100, stop_on_error=0.01, learning_rate=0.01,
+                 is_target=False, ts_n_dims=1, encoder_class=EncoderRNNNumerical):
         super().__init__(is_target)
         self.device, _ = get_devices()
+        self.encoder_class = encoder_class
         self._stop_on_error = stop_on_error
         self._learning_rate = learning_rate
         self._encoded_vector_size = encoded_vector_size
-        self._train_iters = train_iters  # training epochs
+        self._transformer_hidden_size = None
+        self._epochs = train_iters  # training epochs
+        self._pytorch_wrapper = torch.FloatTensor
         self._prepared = False
         self._is_setup = False
         self._max_ts_length = 0
@@ -30,7 +32,6 @@ class RnnEncoder(BaseEncoder):
         self._n_dims = ts_n_dims
         self._normalizer = None
         self._target_ar_normalizers = []
-        self._criterion = nn.MSELoss()
 
     def setup_nn(self, additional_targets=None):
         """This method must be executed after initializing, else self.secondary_type is unassigned"""
@@ -41,6 +42,8 @@ class RnnEncoder(BaseEncoder):
             self._normalizer = MinMaxNormalizer()
 
         total_dims = self._n_dims
+        dec_hsize = self._encoded_vector_size
+
         if additional_targets:
             for t in additional_targets:
                 if t['original_type'] == 'categorical':
@@ -52,8 +55,25 @@ class RnnEncoder(BaseEncoder):
                     t['normalizer'].prepare(t['data'])
                     total_dims += 1
 
-        self._encoder = EncoderRNNNumerical(input_size=total_dims, hidden_size=self._encoded_vector_size).to(self.device)
-        self._decoder = DecoderRNNNumerical(output_size=total_dims, hidden_size=self._encoded_vector_size).to(self.device)
+        if self.encoder_class == EncoderRNNNumerical:
+            self._enc_criterion = nn.MSELoss()
+            self._dec_criterion = self._enc_criterion
+            self._encoder = self.encoder_class(input_size=total_dims,
+                                               hidden_size=self._encoded_vector_size).to(self.device)
+
+        elif self.encoder_class == TransformerEncoder:
+            self._enc_criterion = self._masked_criterion
+            self._dec_criterion = nn.MSELoss()
+            self._base_criterion = nn.MSELoss(reduction="none")
+            if self._transformer_hidden_size is None:
+                self._transformer_hidden_size = total_dims*2  # arbitrary
+
+            self._encoder = self.encoder_class(ninp=total_dims,
+                                               nhead=gcd(dec_hsize, total_dims),
+                                               nhid=self._transformer_hidden_size,
+                                               nlayers=1).to(self.device)
+
+        self._decoder = DecoderRNNNumerical(output_size=total_dims, hidden_size=dec_hsize).to(self.device)
         self._parameters = list(self._encoder.parameters()) + list(self._decoder.parameters())
         self._optimizer = optim.AdamW(self._parameters, lr=self._learning_rate, weight_decay=1e-4)
         self._is_setup = True
@@ -65,6 +85,30 @@ class RnnEncoder(BaseEncoder):
             self._decoder = self._decoder.to(self.device)
         return self
 
+    def _prepare_raw_data(self, data):
+        # Convert to array and determine max length:
+        # priming_data is a list of len B with lists of length in [0, T]
+        # If the length of the series varies a lot, it would be worth it to order them and apply a few optimisations
+        # we do not do so now
+        out_data = []
+        for e in data:
+            if not isinstance(e, torch.Tensor):
+                t = torch.tensor(e, dtype=torch.float, device=self.device)
+            else:
+                t = e.float()
+            t[torch.isnan(t)] = 0.0
+            out_data.append(t)
+        lengths = torch.tensor(
+            [len(e) for e in data], dtype=torch.float, device=self.device
+        )
+        return out_data, lengths
+
+    def _get_batch(self, source, start, step):
+        # source is an iterable element, we want to get source[i+step] or source[i+end]
+        # If padding is not None, until size `source[i+step]`
+        end = min(start + step, len(source))
+        return source[start:end]
+
     def prepare(self, priming_data, previous_target_data=None, feedback_hoop_function=None, batch_size=256):
         """
         The usual, run this on the initial training data for the encoder
@@ -79,107 +123,76 @@ class RnnEncoder(BaseEncoder):
         else:
             self.setup_nn(previous_target_data)
 
-        # Convert to array and determine max length:
-        for i in range(len(priming_data)):
-            if not isinstance(priming_data[i][0], list):
-                priming_data[i] = [priming_data[i]]  # add dimension for 1D timeseries
-            self._max_ts_length = max(len(priming_data[i][0]), self._max_ts_length)  # TODO: this is set at Native... should we still check?
+        # Convert to array and determine max length
+        priming_data, lengths_data = self._prepare_raw_data(priming_data)
+        self._max_ts_length = int(lengths_data.max())
 
-        # normalize data
         if self._normalizer:
             self._normalizer.prepare(priming_data)
+            priming_data = torch.stack([self._normalizer.encode(d) for d in priming_data]).to(self.device)
+        else:
+            priming_data = torch.stack([d for d in priming_data]).unsqueeze(-1).to(self.device)
 
+        # merge all normalized data into a training batch
         if previous_target_data is not None and len(previous_target_data) > 0:
+            normalized_tensors = []
             for target_dict in previous_target_data:
                 normalizer = target_dict['normalizer']
-                target_dict['encoded_data'] = normalizer.encode(target_dict['data'])
                 self._target_ar_normalizers.append(normalizer)
+                data = torch.Tensor(normalizer.encode(target_dict['data'])).to(self.device)
+                data[torch.isnan(data)] = 0.0
+                if len(data.shape) < 3:
+                    data = data.unsqueeze(-1)  # add feature dimension
+                normalized_tensors.append(data)
 
-        # decrease batch_size for small datasets
-        if batch_size >= len(priming_data):
-            batch_size = len(priming_data) // 2
+            normalized_data = torch.cat(normalized_tensors, dim=-1)
+            priming_data = torch.cat([priming_data, normalized_data], dim=-1)
 
         self._encoder.train()
-        for i in range(self._train_iters):
+        for i in range(self._epochs):
             average_loss = 0
-            data_idx = 0
-
-            while data_idx < len(priming_data):
-
-                # batch building, shape: (batch_size, timesteps, n_dims)
-                data_points = priming_data[data_idx:min(data_idx + batch_size, len(priming_data))]
-                batch = []
-                for dp in data_points:
-                    data_tensor = tensor_from_series(dp, self.device, self._n_dims,
-                                                     self._eos, self._max_ts_length,
-                                                     self._normalizer)
-                    batch.append(data_tensor)
-                batch = torch.cat(batch, dim=0).to(self.device)
-
-                # include autoregressive target data
-                if previous_target_data is not None and len(previous_target_data) > 0:
-                    for target_dict in previous_target_data:
-                        t_dp = target_dict['encoded_data'][data_idx:min(data_idx + batch_size, len(priming_data))]
-                        target_tensor = torch.Tensor(t_dp).to(self.device)
-                        target_tensor[torch.isnan(target_tensor)] = 0.0
-                        if len(t_dp.shape) < 3:
-                            target_tensor = target_tensor.unsqueeze(2)
-
-                        # concatenate descriptors
-                        batch = torch.cat((batch, target_tensor), dim=-1)
 
+            for batch_idx in range(0, len(priming_data), batch_size):
                 # setup loss and optimizer
                 self._optimizer.zero_grad()
-                data_idx += batch_size
-                steps = batch.shape[1]
                 loss = 0
 
+                # shape: (batch_size, timesteps, n_dims)
+                batch = self._get_batch(priming_data, batch_idx, min(batch_idx + batch_size, len(priming_data)))
+
+                # encode and decode through time
                 with LightwoodAutocast():
-                    # encode
-                    encoder_hidden = self._encoder.initHidden(self.device, batch_size=batch.shape[0])
-                    next_tensor = batch[:, 0, :].unsqueeze(dim=1)  # initial input
+                    if self.encoder_class == TransformerEncoder:
+                        # pack batch length info tensor
+                        len_batch = self._get_batch(lengths_data, batch_idx, min(batch_idx + batch_size, len(priming_data)))
+                        batch = batch, len_batch
 
-                    for tensor_i in range(steps - 1):
-                        rand = np.random.randint(2)
-                        # teach from forward as well as from known tensor alternatively
-                        if rand == 1:
-                            next_tensor, encoder_hidden = self._encoder.forward(
-                                batch[:, tensor_i, :].unsqueeze(dim=1),
-                                encoder_hidden)
-                        else:
-                            next_tensor, encoder_hidden = self._encoder.forward(next_tensor.detach(), encoder_hidden)
+                        next_tensor, hidden_state, dec_loss = self._encoder.bptt(batch, self._enc_criterion, self.device)
+                        loss += dec_loss
 
-                        loss += self._criterion(next_tensor, batch[:, tensor_i + 1, :].unsqueeze(dim=1))
+                    else:
+                        next_tensor, hidden_state, enc_loss = self._encoder.bptt(batch, self._enc_criterion, self.device)
+                        loss += enc_loss
 
-                    # decode
-                    decoder_hidden = encoder_hidden
-                    next_tensor = torch.full_like(next_tensor, self._sos, dtype=torch.float32).to(self.device)
-                    tensor_target = torch.cat([next_tensor, batch], dim=1)  # add SOS token at t=0 to true input
+                        next_tensor, hidden_state, dec_loss = self._decoder.decode(batch, next_tensor, self._dec_criterion,
+                                                                                   self.device,
+                                                                                   hidden_state=hidden_state)
+                        loss += dec_loss
 
-                    for tensor_i in range(steps - 1):
-                        rand = np.random.randint(2)
-                        # teach from forward as well as from known tensor alternatively
-                        if rand == 1:
-                            next_tensor, decoder_hidden = self._decoder.forward(
-                                tensor_target[:, tensor_i, :].unsqueeze(dim=1),
-                                decoder_hidden)
-                        else:
-                            next_tensor, decoder_hidden = self._decoder.forward(next_tensor.detach(), decoder_hidden)
-
-                        loss += self._criterion(next_tensor, tensor_target[:, tensor_i + 1, :].unsqueeze(dim=1))
-
-                average_loss += loss.item()
                 loss.backward()
+
                 self._optimizer.step()
+                average_loss += loss.item()
 
             average_loss = average_loss / len(priming_data)
+            batch_idx += batch_size
 
             if average_loss < self._stop_on_error:
                 break
             if feedback_hoop_function is not None:
                 feedback_hoop_function("epoch [{epoch_n}/{total}] average_loss = {average_loss}".format(
                     epoch_n=i + 1,
-                    total=self._train_iters,
+                    total=self._epochs,
                     average_loss=average_loss))
 
         self._prepared = True
@@ -196,25 +209,43 @@ class RnnEncoder(BaseEncoder):
         """
         self._encoder.eval()
         with torch.no_grad():
-            # n_timesteps inferred from query
-            data_tensor = tensor_from_series(data, self.device, self._n_dims,
-                                             self._eos, normalizer=self._normalizer)
+            # Convert to array and determine max length
+            data, lengths_data = self._prepare_raw_data(data)
+            _max_ts_length = int(lengths_data.max())
+
+            if self._normalizer:
+                data = torch.stack([self._normalizer.encode(d) for d in data]).to(self.device)
+            else:
+                data = torch.stack([d for d in data]).unsqueeze(-1).to(self.device)
 
             if previous is not None:
-                target_tensor = torch.Tensor(previous).to(self.device)
+                target_tensor = torch.stack(previous).to(self.device)
                 target_tensor[torch.isnan(target_tensor)] = 0.0
                 if len(target_tensor.shape) < 3:
                     target_tensor = target_tensor.transpose(0, 1).unsqueeze(0)
-                data_tensor = torch.cat((data_tensor, target_tensor), dim=-1)
+                data_tensor = torch.cat((data, target_tensor), dim=-1)
+            else:
+                data_tensor = data
 
             steps = data_tensor.shape[1]
-            encoder_hidden = self._encoder.initHidden(self.device)
-            encoder_hidden = encoder_hidden if initial_hidden is None else initial_hidden
 
-            next_tensor = None
-            for tensor_i in range(steps):
-                next_tensor, encoder_hidden = self._encoder.forward(data_tensor[:, tensor_i, :].unsqueeze(dim=0),
-                                                                    encoder_hidden)
+            if self.encoder_class == EncoderRNNNumerical:
+                encoder_hidden = self._encoder.init_hidden(self.device)
+                encoder_hidden = encoder_hidden if initial_hidden is None else initial_hidden
+
+                next_tensor = None
+                for tensor_i in range(steps):
+                    next_tensor, encoder_hidden = self._encoder.forward(data_tensor[:, tensor_i, :].unsqueeze(dim=0),
+                                                                        encoder_hidden)
+            else:
+                next_tensor = None
+                len_batch = self._get_batch(lengths_data, 0, len(data))
+                batch_size, timesteps, _ = data_tensor.shape
+
+                for start_chunk in range(0, timesteps, timesteps):
+                    data, targets, lengths_chunk = get_chunk(data_tensor, len_batch, start_chunk, timesteps)
+                    data = data.transpose(0, 1)
+                    next_tensor, encoder_hidden = self._encoder.forward(data, lengths_chunk, self.device)
 
         if return_next_value:
             return encoder_hidden, next_tensor
@@ -324,3 +355,23 @@ class RnnEncoder(BaseEncoder):
             ret.append(reconstruction)
 
         return torch.Tensor(ret)
+
+    def _masked_criterion(self, output, targets, lengths):
+        """ Computes the loss of the first `lengths` items in the chunk """
+        # Put in (B, T) format and zero-out the unnecessary values
+        mask = len_to_mask(lengths, zeros=False).t()
+
+        # Inflate to feature dimension
+        mask = mask.unsqueeze(-1).repeat(1, 1, output.shape[-1])
+        output = output * mask
+        targets = targets * mask
+
+        # compute the loss with respect to the appropriate lengths and average across the batch-size
+        # We compute for every output (x_i)_i=1^L and target (y_i)_i=1^L, loss = 1/L \sum (x_i - y_i)^2
+        # And average across the mini-batch
+        losses = self._base_criterion(output, targets).sum(dim=2).sum(dim=0)
+
+        # The TBPTT will compute a slightly different loss, but it is not problematic
+        loss = torch.dot((1.0 / lengths.float()), losses) / len(losses)
+
+        return loss
