@@ -187,14 +187,14 @@ class NnMixer(BaseMixer):
 
                     if output_type == COLUMN_DATA_TYPES.CATEGORICAL:
                         self.criterion_arr.append(TransformCrossEntropyLoss(weight=weights_slice))
-                        self.unreduced_criterion_arr.append(TransformCrossEntropyLoss(weight=weights_slice, reduce=False))
+                        self.unreduced_criterion_arr.append(TransformCrossEntropyLoss(weight=weights_slice, reduction='none'))
                     elif output_type == COLUMN_DATA_TYPES.MULTIPLE_CATEGORICAL:
                         self.criterion_arr.append(torch.nn.BCEWithLogitsLoss(weight=weights_slice))
-                        self.unreduced_criterion_arr.append(torch.nn.BCEWithLogitsLoss(weight=weights_slice, reduce=False))
+                        self.unreduced_criterion_arr.append(torch.nn.BCEWithLogitsLoss(weight=weights_slice, reduction='none'))
                 # Note: MSELoss works great for numeric, for the other types it's more of a placeholder
                 else:
                     self.criterion_arr.append(torch.nn.MSELoss())
-                    self.unreduced_criterion_arr.append(torch.nn.MSELoss(reduce=False))
+                    self.unreduced_criterion_arr.append(torch.nn.MSELoss(reduction='none'))
 
         self.optimizer_class = Ranger
         if self.optimizer_args is None:
@@ -336,8 +336,8 @@ class NnMixer(BaseMixer):
 
                         last_test_error = test_error
 
-                        delta_mean = np.mean(test_error_delta_buff[-5:])
-                        subset_delta_mean = np.mean(subset_test_error_delta_buff[-5:])
+                        delta_mean = np.mean(test_error_delta_buff[-5:]) if test_error_delta_buff else 0
+                        subset_delta_mean = np.mean(subset_test_error_delta_buff[-5:]) if subset_test_error_delta_buff else 0
 
                         if self._nonpersistent['callback'] is not None:
                             self._nonpersistent['callback'](epoch, training_error, test_error, delta_mean, self.calculate_accuracy(test_ds))
@@ -684,6 +684,12 @@ class NnMixer(BaseMixer):
 
             if 'weights' in ds.get_column_config(output_column):
                 weight_map = ds.get_column_config(output_column)['weights']
+                # omit points for which there's no target info in timeseries forecasting
+                if '_timestep_' in ds.get_column_config(output_column)['name']:
+                    classes = set(weight_map.keys())
+                    observed_classes = set(reals)
+                    for unrecognized_class in observed_classes.difference(classes):
+                        weight_map[unrecognized_class] = 0
             else:
                 weight_map = None
 
